local function nilzero(n)
	if n == 0 then
		return nil
	end
	return n
end

local createBlueprint

local bpMeta = {
	__index = {
		DumpMaterialExpenses = function(self, outFile, priceTable, formatNumber)
			local cost = 0
			for item,quantity in pairs(self.materials) do
				cost = cost + priceTable[item] * quantity
				outFile:write("<tr><td>"..item.."</td>"
					.."<td>"..formatNumber(priceTable[item]).."</td>"
					.."<td>x</td>"
					.."<td>"..formatNumber(quantity).."</td>"
					.."<td>=</td>"
					.."<td>"..formatNumber(priceTable[item] * quantity).."</td>"
					.."<td>...</td><td>"..formatNumber(cost).."</td></tr>")
			end
			for item,damage in pairs(self.tools) do
				cost = cost + priceTable[item] * damage
				outFile:write("<tr><td>"..item.."</td>"
					.."<td>"..formatNumber(priceTable[item]).."</td>"
					.."<td>x</td>"
					.."<td>"..formatNumber(damage).."</td>"
					.."<td>=</td>"
					.."<td>"..formatNumber(priceTable[item] * damage).."</td>"
					.."<td>...</td><td>"..formatNumber(cost).."</td></tr>")
			end
			outFile:write("<tr><td colspan='4'></td><td>Total:</td><td>"..formatNumber(cost).."</td></tr>")
			return cost
		end,
		ConstructionCost = function(self, priceTable)
			local cost = 0
			for item,quantity in pairs(self.materials) do
				assert(priceTable[item], "No price for item: "..item)
				cost = cost + priceTable[item] * quantity
			end
			for item,damage in pairs(self.tools) do
				assert(priceTable[item], "No price for tool: "..item)
				cost = cost + priceTable[item] * damage
			end
			return cost
		end,
		ComponentConstructionTime = function(self, bpdb)
			local totalTime = 0
			for item,quantity in pairs(self.materials) do
				if bpdb[item] then
					totalTime = totalTime + bpdb[item].buildTime * quantity
				end
			end
			for item,damage in pairs(self.tools) do
				if bpdb[item] then
					totalTime = totalTime + bpdb[item].buildTime * damage
				end
			end
			return totalTime
		end,
		DumpDatacoreExpenses = function(self, outFile, priceTable, formatNumber, params)
			local chance = params.inventedBP.chance
			local t2runs = params.inventedBP.runs
			local decryptor = params.parameters.decryptor.item

			local cost = 0
			for _,coreItem in pairs(self.inventionDatacores) do
				cost = cost + priceTable[coreItem.item] * coreItem.quantity
				outFile:write("<tr><td>"..coreItem.item.."</td>"
					.."<td>"..formatNumber(priceTable[coreItem.item]).."</td>"
					.."<td>x</td>"
					.."<td>"..formatNumber(coreItem.quantity).."</td>"
					.."<td>=</td>"
					.."<td>"..formatNumber(priceTable[coreItem.item] * coreItem.quantity).."</td>"
					.."<td>...</td><td>"..formatNumber(cost).."</td></tr>")
			end

			if decryptor ~= "No Decryptor" then
				cost = cost + priceTable[decryptor]
				outFile:write("<tr><td>"..decryptor.."</td>"
					.."<td>"..formatNumber(priceTable[decryptor]).."</td>"
					.."<td>x</td>"
					.."<td>1</td>"
					.."<td>=</td>"
					.."<td>"..formatNumber(priceTable[decryptor]).."</td>"
					.."<td>...</td><td>"..formatNumber(cost).."</td></tr>")
			end

			outFile:write("<tr><td colspan='4'></td><td>Total:</td><td>"..formatNumber(cost).."</td></tr>")
			cost = cost / chance
			outFile:write("<tr><td colspan='4'></td><td>/ "..chance.." chance =</td><td>"..formatNumber(cost).."</td></tr>")

			if t2runs ~= 1 then
				cost = cost / t2runs
				outFile:write("<tr><td colspan='4'></td><td>/ "..t2runs.." runs =</td><td>"..formatNumber(cost).."</td></tr>")
			end
		end,
		InventionCost = function(self, priceTable)
			local cost = 0
			for _,coreItem in pairs(self.inventionDatacores) do
				cost = cost + priceTable[coreItem.item] * coreItem.quantity
			end
			return cost
		end,
		InventionRuns = function(self, params)
			assert(params.t1bp.maxRuns, "No max runs for T1 blueprint "..params.t1bp.item);
			assert(self.maxRuns, "No max runs for T2 blueprint "..self.item);

			local baseRuns = math.floor(
					(
						 ( params.baseRuns / params.t1bp.maxRuns ) * self.maxRuns / 10
					) + params.decryptor.runsModifier)
			return max( 1, baseRuns	)
		end,
		Invent = function(self, params)
			local resultRuns = self:InventionRuns(params)
			local resultME = params.decryptor.meModifier - 4
			local resultPE = params.decryptor.peModifier - 4
			local resultSuccess = self.inventionBaseChance *
				(1 + (0.01 * params.encryptionSkill)) *
				(1 + ((params.datacoreSkills) * (0.1 / (5 - params.metaItemLevel)))) *
				params.decryptor.probModifier

			local bp = createBlueprint(self.item, self.materials)
			bp.me = resultME
			bp.pe = resultPE
			bp.tools = self.tools
			bp.chance = resultSuccess
			bp.runs = resultRuns
			bp.buildTime = self.buildTime
			bp:ApplyWaste()

			return bp;
		end,
		ApplyWaste = function(self)
			local newMaterials = { }
			local matMultiplier = 1
			local timeMultiplier = 1

			if self.me <= 0 then
				matMultiplier = 1 + 0.1 * (1 - self.me)
			else
				matMultiplier = 1 + 0.1 / (1 + self.me)
			end

			if self.pe <= 0 then
				timeMultiplier = 1 + 0.1 * (1 - self.pe)
			else
				timeMultiplier = 1 + 0.1 / (1 + self.pe)
			end

			matMultiplier = matMultiplier * (1 + (0.05 * (5 - config.productionEfficiencySkillLevel)))

			for item,quantity in pairs(self.materials) do
				newMaterials[item] = math.floor(quantity * matMultiplier + 0.5)
			end
			self.materials = newMaterials

			assert(self.buildTime, "No build time defined for "..self.item)
			self.buildTime = self.buildTime * timeMultiplier
		end,
	}
}

createBlueprint = function(item, mats)
	local bp = {
		item = item,
		me = "infinite",
		pe = "infinite",
		materials = mats,
		tools = { }
	}

	setmetatable(bp, bpMeta)
	return bp
end

blueprints =
{
	inventionParameters = function(baseRuns, t1bp, decryptor, encryptionSkill, datacore1Skill, datacore2Skill, metaItemLevel)
		return {
			baseRuns = baseRuns,
			t1bp = t1bp,
			decryptor = decryptor,
			encryptionSkill = encryptionSkill,
			datacoreSkills = datacore1Skill + datacore2Skill,
			metaItemLevel = metaItemLevel
		}
	end,

	load = function()
		local blueprints = { }
		local t1blueprints = loadTabFile("data/cache/blueprints_t1.dat")

		print("Loading BPs...")
		for _,row in ipairs(t1blueprints.rows) do
			local mats = {
				Tritanium = nilzero(row.tritanium),
				Pyerite = nilzero(row.pyerite),
				Mexallon = nilzero(row.mexallon),
				Isogen = nilzero(row.isogen),
				Nocxium = nilzero(row.nocxium),
				Zydrine = nilzero(row.zydrine),
				Megacyte = nilzero(row.megacyte),
			}
			blueprints[row.item] = createBlueprint(row.item, mats)
		end

		local complexBlueprints = loadTabFile("data/cache/blueprints_complex.dat")

		local itemMaterials = { }

		for _,row in ipairs(complexBlueprints.rows) do
			local im = itemMaterials[row.item]
			if im == nil then
				im = { }
				itemMaterials[row.item] = im
			end
			im[row.material] = row.quantity
		end

		for k,v in pairs(itemMaterials) do
			blueprints[k] = createBlueprint(k, v)
		end

		local toolsDB = loadTabFile("data/cache/blueprints_tools.dat")
		database.indexByMultiple(toolsDB, "item")

		for item,toolList in pairs(toolsDB.keyBy.item) do
			for _,toolRow in ipairs(toolList) do
				if blueprints[item] == nil then
					blueprints[item] = createBlueprint(item, { } )
				end
				blueprints[item].tools[toolRow.tool] = toolRow.damage
			end
		end

		local bpDB = loadTabFile("data/cache/blueprints_properties.dat")

		for _,row in pairs(bpDB.rows) do
			if blueprints[row.item] then
				blueprints[row.item].copyTime = row.copyTime
				blueprints[row.item].maxRuns = row.maxRuns
				blueprints[row.item].buildTime = row.buildTime
				blueprints[row.item].inventTime = row.inventTime
			end
		end

		local ibpDB = loadTabFile("data/cache/blueprints_invention.dat")

		for _,row in pairs(ibpDB.rows) do
			assert(blueprints[row.item], "Unknown item in invention blueprints: "..row.item)
			blueprints[row.item].inventedFrom = row.inventedFrom
			blueprints[row.item].inventionBaseChance = row.baseChance
			blueprints[row.item].inventionDatacores = {
				{ item = row.datacore1, quantity = row.datacore1quantity },
				{ item = row.datacore2, quantity = row.datacore2quantity },
			}
		end

		return blueprints
	end,
}
