assert(loadfile("code/tabloader.lua"))()
assert(loadfile("code/blueprints.lua"))()
assert(loadfile("code/dbtools.lua"))()
assert(loadfile("config.lua"))()

local marketDB = loadTabFile("data/items_marketids.dat")

local function unformatNumber(str)
	local result = ""
	local validChars = {
		["0"] = true,
		["1"] = true,
		["2"] = true,
		["3"] = true,
		["4"] = true,
		["5"] = true,
		["6"] = true,
		["7"] = true,
		["8"] = true,
		["9"] = true,
		["."] = true,
	}
	for i=1,#str do
		local c = string.sub(str,i,i)
		if validChars[c] then result = result..c end
	end
	return tonumber(result)
end

local function getItemPriceLevel(sourceText, itemName)
	local sellOrders = string.match(sourceText, '<div id="sell_orders">(.-)</div>')
	local allOrders = { }

	for order in string.gmatch(sellOrders, "<tr bgcolor=.->(.-)</tr>") do
		-- Region, station, price, quantity
		local price, vol = string.match(order, "<td>.-</td>.-<td>.-</td>.-<td>(.-)</td>.-<td>(.-)</td>")
		allOrders[#allOrders+1] = { price = unformatNumber(price), volume = unformatNumber(vol) }
	end

	local merged = { allOrders[1] }
	for i=2,#allOrders do
		local lastOrder = merged[#merged]
		local nextOrder = allOrders[i]

		-- Aggregate prices within 0.1%
		if nextOrder.price < lastOrder.price * 1.001 then
			local combinedVolume = lastOrder.volume + nextOrder.volume
			local combinedValue = lastOrder.volume*lastOrder.price + nextOrder.volume*nextOrder.price
			lastOrder.volume = combinedVolume
			lastOrder.price = combinedValue / combinedVolume
		else
			merged[#merged+1] = nextOrder
		end
	end

	local premerged = allOrders
	allOrders = merged

	table.sort(allOrders, function(a, b) return (a.price < b.price) end)

	-- Total the volume of the first 5 orders
	local total = 0
	local scanDepth = 5
	local tolerance = 0.04

	local throwWarning = false

	if allOrders[scanDepth] == nil then
		scanDepth = #allOrders
		if scanDepth == 0 then
			print("\n!!! Zero entries for "..itemName..", can't compute price level! !!!\nExported as 1 ISK, you should fix this manually,\n   or make a price override in config.lua.\n")
			return 1
		elseif (#premerged < scanDepth) then
			print("\nWARNING: Item "..itemName.." only has "..scanDepth.." entries!\n   Eve-Central might be malfunctioning!")
			local throwWarning = true
		end
	end

	for i=1,scanDepth do
		if allOrders[i] then
			total = total + allOrders[i].volume
		end
	end

	for i=1,scanDepth do
		if allOrders[i].volume / total > tolerance then
			if throwWarning then
				print("Exported price as "..allOrders[i].price.." ISK,\n   if this isn't correct you should fix it manually,\n   or override it in config.lua")
			end
			return allOrders[i].price
		end
	end

end

local function main()
	local priceTable = config.priceOverrides

	for i,row in ipairs(marketDB.rows) do
		if priceTable[row.name] == nil then
			local f = io.open("marketdumps/item"..row.id..".html", "r")
			local fText = f:read("*a")
			f:close()

			local priceLevel = math.floor(getItemPriceLevel(fText, row.name) * 100 + 0.5) / 100

			priceTable[row.name] = priceLevel
		end
	end

	local blueprintsLibrary = blueprints.load()

	-- Compute intermediates
	local intermediateItemsDB = loadTabFile("data/items_intermediates.dat")
	for _,row in ipairs(intermediateItemsDB.rows) do
		local blueprint = blueprintsLibrary[row.item]
		if blueprint == nil then
			priceTable[row.item] = 0
		else
			priceTable[row.item] = blueprint:ConstructionCost(priceTable)
		end
	end

	local mpf = io.open("data/cache/items_marketprices.dat", "w")
	mpf:write("item\tprice")
	for k,v in pairs(priceTable) do
		mpf:write("\n"..k.."\t"..v)
	end
	mpf:close()
end



main()